-- WarlockDemonology.lua
-- June 2018

local addon, ns = ...
local Hekili = _G[ addon ]

local class = Hekili.Class
local state = Hekili.State

local PTR = ns.PTR

local FindUnitBuffByID, FindUnitDebuffByID = ns.FindUnitBuffByID, ns.FindUnitDebuffByID


-- Conduits
-- [-] borne_of_blood
-- [-] carnivorous_stalkers
-- [-] fel_commando
-- [x] tyrants_soul


if UnitClassBase( "player" ) == "WARLOCK" then
    local spec = Hekili:NewSpecialization( 266, true )

    spec:RegisterResource( Enum.PowerType.SoulShards )
    spec:RegisterResource( Enum.PowerType.Mana )

    -- Talents
    spec:RegisterTalents( {
        dreadlash = 19290, -- 264078
        bilescourge_bombers = 22048, -- 267211
        demonic_strength = 23138, -- 267171

        demonic_calling = 22045, -- 205145
        power_siphon = 21694, -- 264130
        doom = 23158, -- 603

        demon_skin = 19280, -- 219272
        burning_rush = 19285, -- 111400
        dark_pact = 19286, -- 108416

        from_the_shadows = 22477, -- 267170
        soul_strike = 22042, -- 264057
        summon_vilefiend = 23160, -- 264119

        darkfury = 22047, -- 264874
        mortal_coil = 19291, -- 6789
        howl_of_terror = 23465, -- 5484

        soul_conduit = 23147, -- 215941
        inner_demons = 23146, -- 267216
        grimoire_felguard = 21717, -- 111898

        sacrificed_souls = 23161, -- 267214
        demonic_consumption = 22479, -- 267215
        nether_portal = 23091, -- 267217
    } )


    -- PvP Talents
    spec:RegisterPvpTalents( { 
        amplify_curse = 3507, -- 328774
        bane_of_fragility = 3505, -- 199954
        call_fel_lord = 162, -- 212459
        call_felhunter = 156, -- 212619
        call_observer = 165, -- 201996
        casting_circle = 3626, -- 221703
        essence_drain = 3625, -- 221711
        fel_obelisk = 5400, -- 353601
        gateway_mastery = 3506, -- 248855
        master_summoner = 1213, -- 212628
        nether_ward = 3624, -- 212295
        pleasure_through_pain = 158, -- 212618
        shadow_rift = 5394, -- 353294
    } )


    -- Demon Handling
    local dreadstalkers = {}
    local dreadstalkers_v = {}

    local vilefiend = {}
    local vilefiend_v = {}

    local wild_imps = {}
    local wild_imps_v = {}

    local demonic_tyrant = {}
    local demonic_tyrant_v = {}

    local grim_felguard = {}
    local grim_felguard_v = {}

    local other_demon = {}
    local other_demon_v = {}

    local imps = {}
    local guldan = {}
    local guldan_v = {}

    local last_summon = {}

    local FindUnitBuffByID = ns.FindUnitBuffByID


    local shards_for_guldan = 0

    local function UpdateShardsForGuldan()
        shards_for_guldan = UnitPower( "player", Enum.PowerType.SoulShards )
    end


    -- tyrant_ready needs to be handled internally, as the priority interpreter can't keep variable state in the same way that simc does.
    local tyrant_ready_actual = false

    spec:RegisterStateExpr( "tyrant_ready", function ()
        return tyrant_ready_actual
    end )

    spec:RegisterStateFunction( "update_tyrant_readiness", function( hog_shards )
        if not IsSpellKnown( 265187 ) then return end

        hog_shards = hog_shards or 1

        -- The last part of a Tyrant Prep phase should be a HoG cast.
        if cooldown.summon_demonic_tyrant.remains < 4 then
            if  ( talent.doom.enabled and not debuff.doom.up ) or
                ( talent.demonic_strength.enabled and not talent.demonic_consumption.enabled and cooldown.demonic_strength.remains < 4 ) or
                ( talent.nether_portal.enabled and cooldown.nether_portal.remains < 4 ) or
                ( talent.grimoire_felguard.enabled and cooldown.grimoire_felguard.remains < 4 ) or
                ( talent.summon_vilefiend.enabled and cooldown.summon_vilefiend.remains < 4 ) or
                ( cooldown.call_dreadstalkers.remains < 4 ) or
                ( buff.demonic_core.up and soul_shard < 4 and ( talent.demonic_consumption.enabled or buff.nether_portal.down ) ) then
                    if Hekili.ActiveDebug then Hekili:Debug( "Flagging 'tyrant_ready' as false due to missed APL conditions." ) end
                    tyrant_ready = false
                    return
            end

            if ( soul_shard + hog_shards >= ( buff.nether_portal.up and 1 or 5 ) ) then
                if Hekili.ActiveDebug then Hekili:Debug( "Flagging 'tyrant_ready' as true as all conditions were met." ) end
                tyrant_ready = true
            end

            if Hekili.ActiveDebug then Hekili:Debug( "Leaving 'tyrant_ready' as " .. tostring( tyrant_ready ) .. "." ) end
            return
        end

        if Hekili.ActiveDebug then Hekili:Debug( "Flagging 'tyrant_ready' as false based on cooldown." ) end
        tyrant_ready = false
    end )


    local hog_time = 0

    spec:RegisterEvent( "COMBAT_LOG_EVENT_UNFILTERED", function()
        local _, subtype, _, source, _, _, _, destGUID, _, _, _, spellID, spellName = CombatLogGetCurrentEventInfo()

        local now = GetTime()

        if source == state.GUID then
            if subtype == "SPELL_SUMMON" then
                -- Dreadstalkers: 104316, 12 seconds uptime.
                -- if spellID == 193332 or spellID == 193331 then table.insert( dreadstalkers, now + 12 )

                -- Vilefiend: 264119, 15 seconds uptime.
                -- elseif spellID == 264119 then table.insert( vilefiend, now + 15 )

                -- Wild Imp: 104317 and 279910, 20 seconds uptime.
                -- else
                if spellID == 104317 or spellID == 279910 then
                    table.insert( wild_imps, now + 20 )

                    imps[ destGUID ] = {
                        t = now,
                        casts = 0,
                        expires = math.ceil( now + 20 ),
                        max = math.ceil( now + 20 )
                    }

                    if guldan[ 1 ] then
                        -- If this imp is impacting within 0.15s of the expected queued imp, remove that imp from the queue.
                        if abs( now - guldan[ 1 ] ) < 0.15 then
                            table.remove( guldan, 1 )
                        end
                    end

                    -- Expire missed/lost Gul'dan predictions.
                    while( guldan[ 1 ] ) do
                        if guldan[ 1 ] < now then
                            table.remove( guldan, 1 )
                        else
                            break
                        end
                    end

                -- Grimoire Felguard
                -- elseif spellID == 111898 then table.insert( grim_felguard, now + 17 )

                -- Demonic Tyrant: 265187, 15 seconds uptime.
                elseif spellID == 265187 then table.insert( demonic_tyrant, now + 15 )
                    -- for i = 1, #dreadstalkers do dreadstalkers[ i ] = dreadstalkers[ i ] + 15 end
                    -- for i = 1, #vilefiend do vilefiend[ i ] = vilefiend[ i ] + 15 end
                    -- for i = 1, #grim_felguard do grim_felguard[ i ] = grim_felguard[ i ] + 15 end
                    for i = 1, #wild_imps do wild_imps[ i ] = wild_imps[ i ] + 15 end

                    for _, imp in pairs( imps ) do
                        imp.expires = imp.expires + 15
                        imp.max = imp.max + 15
                    end


                -- Other Demons, 15 seconds uptime.
                -- 267986 - Prince Malchezaar
                -- 267987 - Illidari Satyr
                -- 267988 - Vicious Hellhound
                -- 267989 - Eyes of Gul'dan
                -- 267991 - Void Terror
                -- 267992 - Bilescourge
                -- 267994 - Shivarra
                -- 267995 - Wrathguard
                -- 267996 - Darkhound
                elseif spellID >= 267986 and spellID <= 267996 then table.insert( other_demon, now + 15 ) end

            elseif subtype == "SPELL_CAST_START" and spellID == 105174 then
                C_Timer.After( 0.25, UpdateShardsForGuldan )

            elseif subtype == "SPELL_CAST_SUCCESS" then
                -- Implosion.
                if spellID == 196277 then
                    table.wipe( wild_imps )
                    table.wipe( imps )

                -- Power Siphon.
                elseif spellID == 264130 then
                    if wild_imps[1] then table.remove( wild_imps, 1 ) end
                    if wild_imps[1] then table.remove( wild_imps, 1 ) end

                    for i = 1, 2 do
                        local lowest

                        for id, imp in pairs( imps ) do
                            if not lowest then lowest = id
                            elseif imp.expires < imps[ lowest ].expires then
                                lowest = id
                            end
                        end

                        if lowest then
                            imps[ lowest ] = nil
                        end
                    end

                -- Hand of Guldan (queue imps).
                elseif spellID == 105174 then
                    hog_time = GetTime()

                    if shards_for_guldan >= 1 then table.insert( guldan, now + 0.6 ) end
                    if shards_for_guldan >= 2 then table.insert( guldan, now + 0.8 ) end
                    if shards_for_guldan >= 3 then table.insert( guldan, now + 1 ) end

                    -- Per SimC APL, we go into Tyrant with 5 shards -OR- with Nether Portal up.
                    if IsSpellKnown( 265187 ) then
                        local start, duration = GetSpellCooldown( 265187 )

                        if start and duration and start + duration - now < 4 then
                            state.reset()

                            local np = state.talent.nether_portal.enabled and FindUnitBuffByID( "player", 267218 )

                            if ( not state.talent.doom.enabled or state.action.doom.lastCast - now < 30 ) and 
                            ( state.cooldown.demonic_strength.remains > 4 or not state.talent.demonic_strength.enabled or state.talent.demonic_consumption.enabled ) and 
                            ( state.cooldown.nether_portal.remains > 4 or not state.talent.nether_portal.enabled ) and
                            ( state.cooldown.grimoire_felguard.remains > 4 or not state.talent.grimoire_felguard.enabled ) and 
                            ( state.cooldown.summon_vilefiend.remains > 4 or not state.talent.summon_vilefiend.enabled ) and
                            ( state.cooldown.call_dreadstalkers.remains > 4 ) and
                            ( state.buff.demonic_core.down or shards_for_guldan > 3 or ( not state.talent.demonic_consumption.enabled or np ) ) and
                            ( shards_for_guldan == 5 or np ) then
                                tyrant_ready_actual = true
                            end
                        else
                            tyrant_ready_actual = false
                        end
                    end

                -- Summon Demonic Tyrant.
                elseif spellID == 265187 then
                    tyrant_ready_actual = false

                end
            end
        
        elseif imps[ source ] and subtype == "SPELL_CAST_START" then
            local demonic_power = FindUnitBuffByID( "player", 265273 )

            if not demonic_power then
                local imp = imps[ source ]

                imp.start = now
                imp.casts = imp.casts + 1

                imp.expires = min( imp.max, now + ( ( ( state.level > 55 and 7 or 6 ) - imp.casts ) * 2 * state.haste ) )
            end
        end
    end )


    local wipe = table.wipe

    spec:RegisterHook( "reset_precast", function()
        local i = 1

        for id, imp in pairs( imps ) do
            if imp.expires < now then
                imps[ id ] = nil
            end
        end

        while( wild_imps[ i ] ) do
            if wild_imps[ i ] < now then
                table.remove( wild_imps, i )
            else
                i = i + 1
            end
        end

        wipe( wild_imps_v )
        for n, t in pairs( imps ) do table.insert( wild_imps_v, t.expires ) end
        table.sort( wild_imps_v )

        local difference = #wild_imps_v - GetSpellCount( 196277 )

        while difference > 0 do
            table.remove( wild_imps_v, 1 )
            difference = difference - 1
        end

        wipe( guldan_v )
        for n, t in ipairs( guldan ) do guldan_v[ n ] = t end


        i = 1
        while( other_demon[ i ] ) do
            if other_demon[ i ] < now then
                table.remove( other_demon, i )
            else
                i = i + 1
            end
        end

        wipe( other_demon_v )
        for n, t in ipairs( other_demon ) do other_demon_v[ n ] = t end


        if #dreadstalkers_v > 0  then wipe( dreadstalkers_v ) end
        if #vilefiend_v > 0      then wipe( vilefiend_v )     end
        if #grim_felguard_v > 0  then wipe( grim_felguard_v ) end
        if #demonic_tyrant_v > 0 then wipe( demonic_tyrant_v ) end

        -- Pull major demons from Totem API.
        for i = 1, 5 do
            local exists, name, summoned, duration, texture = GetTotemInfo( i )

            if exists then
                local demon

                -- Grimoire Felguard
                if texture == 136216 then demon = grim_felguard_v
                elseif texture == 1616211 then demon = vilefiend_v
                elseif texture == 1378282 then demon = dreadstalkers_v
                elseif texture == 135002 then demon = demonic_tyrant_v end

                if demon then
                    insert( demon, summoned + duration )
                end
            end

            if #grim_felguard_v > 1 then table.sort( grim_felguard_v ) end
            if #vilefiend_v > 1 then table.sort( vilefiend_v ) end
            if #dreadstalkers_v > 1 then table.sort( dreadstalkers_v ) end
            if #demonic_tyrant_v > 1 then table.sort( demonic_tyrant_v ) end
        end

        last_summon.name = nil
        last_summon.at = nil
        last_summon.count = nil

        if demonic_tyrant_v[ 1 ] and demonic_tyrant_v[ 1 ] > query_time then
            summonPet( "demonic_tyrant", demonic_tyrant_v[ 1 ] - query_time )
        end

        if buff.demonic_power.up and buff.demonic_power.remains > pet.demonic_tyrant.remains then
            summonPet( "demonic_tyrant", buff.demonic_power.remains )
        end

        tyrant_ready = nil

        if cooldown.summon_demonic_tyrant.remains > 5 then
            tyrant_ready = false
        end

        local subjugated, icon, count, debuffType, duration, expirationTime = FindUnitDebuffByID( "pet", 1098 )
        if subjugated then
            summonPet( "subjugated_demon", expirationTime - now )
        else
            dismissPet( "subjugated_demon" )
        end

        if Hekili.ActiveDebug then
            Hekili:Debug(   "Is Tyrant Ready?: %s\n" ..
                            " - Dreadstalkers: %d, %.2f\n" ..
                            " - Vilefiend    : %d, %.2f\n" ..
                            " - Grim Felguard: %d, %.2f\n" ..
                            " - Wild Imps    : %d, %.2f\n" ..
                            "Next Demon Exp. : %.2f",
                            tyrant_ready and "Yes" or "No",
                            buff.dreadstalkers.stack, buff.dreadstalkers.remains,
                            buff.vilefiend.stack, buff.vilefiend.remains,
                            buff.grimoire_felguard.stack, buff.grimoire_felguard.remains,
                            buff.wild_imps.stack, buff.wild_imps.remains,
                            major_demon_expires )
        end
    end )


    spec:RegisterHook( "advance_end", function ()
        for i = #guldan_v, 1, -1 do
            local imp = guldan_v[i]

            if imp <= query_time then
                if ( imp + 20 ) > query_time then
                    insert( wild_imps_v, imp + 20 )
                end
                remove( guldan_v, i )
            end
        end
    end )


    -- Provide a way to confirm if all Hand of Gul'dan imps have landed.
    spec:RegisterStateExpr( "spawn_remains", function ()
        if #guldan_v > 0 then
            return max( 0, guldan_v[ #guldan_v ] - query_time )
        end
        return 0
    end )


    spec:RegisterHook( "spend", function( amt, resource )
        if resource == "soul_shards" and amt > 0 then
            if buff.nether_portal.up then
                summon_demon( "other", 15, amt )
            end

            if legendary.wilfreds_sigil_of_superior_summoning.enabled then
                reduceCooldown( "summon_demonic_tyrant", amt * 0.6 )
            end
        end
    end )


    spec:RegisterStateFunction( "summon_demon", function( name, duration, count )
        local db = other_demon_v

        if name == "dreadstalkers" then db = dreadstalkers_v
        elseif name == "vilefiend" then db = vilefiend_v
        elseif name == "wild_imps" then db = wild_imps_v
        elseif name == "grimoire_felguard" then db = grim_felguard_v
        elseif name == "demonic_tyrant" then db = demonic_tyrant_v end

        count = count or 1
        local expires = query_time + duration

        last_summon.name = name
        last_summon.at = query_time
        last_summon.count = count

        for i = 1, count do
            table.insert( db, expires )
        end
    end )


    spec:RegisterStateFunction( "extend_demons", function( duration )
        duration = duration or 15

        for k, v in pairs( dreadstalkers_v ) do dreadstalkers_v[ k ] = v + duration end
        for k, v in pairs( vilefiend_v     ) do vilefiend_v    [ k ] = v + duration end
        for k, v in pairs( wild_imps_v     ) do wild_imps_v    [ k ] = v + duration end
        for k, v in pairs( grim_felguard_v ) do grim_felguard_v[ k ] = v + duration end
        for k, v in pairs( other_demon_v   ) do other_demon_v  [ k ] = v + duration end
    end )


    spec:RegisterStateFunction( "consume_demons", function( name, count )
        local db = other_demon_v

        if name == "dreadstalkers" then db = dreadstalkers_v
        elseif name == "vilefiend" then db = vilefiend_v
        elseif name == "wild_imps" then db = wild_imps_v
        elseif name == "grimoire_felguard" then db = grim_felguard_v
        elseif name == "demonic_tyrant" then db = demonic_tyrant_v end

        if type( count ) == "string" and count == "all" then
            table.wipe( db )

            -- Wipe queued Guldan imps that should have landed by now.
            if name == "wild_imps" then
                while( guldan_v[ 1 ] ) do
                    if guldan_v[ 1 ] < now then table.remove( guldan_v, 1 )
                    else break end
                end
            end
            return
        end

        count = count or 0

        if count >= #db then
            count = count - #db
            table.wipe( db )
        end

        while( count > 0 ) do
            if not db[1] then break end
            table.remove( db, 1 )
            count = count - 1
        end

        if name == "wild_imps" and count > 0 then
            while( count > 0 ) do
                if not guldan_v[1] or guldan_v[1] > now then break end
                table.remove( guldan_v, 1 )
                count = count - 1
            end
        end
    end )


    spec:RegisterStateExpr( "soul_shard", function () return soul_shards.current end )

    -- How long before you can complete a 3 Soul Shard HoG cast.
    spec:RegisterStateExpr( "time_to_hog", function ()
        local shards_needed = max( 0, 3 - soul_shards.current )
        local cast_time = action.hand_of_guldan.cast_time

        if shards_needed > 0 then
            local cores = min( shards_needed, buff.demonic_core.stack )

            if cores > 0 then
                cast_time = cast_time + cores * gcd.execute
                shards_needed = shards_needed - cores
            end

            cast_time = cast_time + shards_needed * action.shadow_bolt.cast_time
        end

        return cast_time
    end )


    spec:RegisterStateExpr( "major_demons_active", function ()
        return ( buff.grimoire_felguard.up and 1 or 0 ) + ( buff.vilefiend.up and 1 or 0 ) + ( buff.dreadstalkers.up and 1 or 0 )
    end )


    -- When the next major demon (anything but Wild Imps) expires.
    spec:RegisterStateExpr( "major_demon_expires", function ()
        local expire = 3600

        if buff.grimoire_felguard.up then expire = min( expire, buff.grimoire_felguard.remains ) end
        if buff.vilefiend.up then expire = min( expire, buff.vilefiend.remains ) end
        if buff.dreadstalkers.up then expire = min( expire, buff.dreadstalkers.remains ) end

        if expire == 3600 then return 0 end
        return expire
    end )


    -- New imp forecasting expressions for Demo.
    spec:RegisterStateExpr( "incoming_imps", function ()
        local n = 0

        for i, time in ipairs( guldan_v ) do
            if time < query_time then break end
            n = n + 1
        end

        return n
    end )


    local time_to_n = 0

    spec:RegisterStateTable( "query_imp_spawn", setmetatable( {}, {
        __index = function( t, k )
            if k ~= "remains" then return 0 end

            local queued = #guldan_v

            if queued == 0 then return 0 end

            if time_to_n == 0 or time_to_n >= queued then
                return max( 0, guldan_v[ queued ] - query_time )
            end

            local count = 0
            local remains = 0

            for i, time in ipairs( guldan_v ) do
                if time > query_time then
                    count = count + 1
                    remains = time - query_time

                    if count >= time_to_n then break end
                end
            end

            return remains
        end,
    } ) )

    spec:RegisterStateTable( "time_to_imps", setmetatable( {}, {
        __index = function( t, k )
            if type( k ) == "number" then
                time_to_n = min( #guldan_v, k )
            elseif k == "all" then
                time_to_n = #guldan_v
            else
                time_to_n = 0
            end

            return query_imp_spawn
        end
    } ) )

    local debugstack = debugstack

    spec:RegisterStateTable( "imps_spawned_during", setmetatable( {}, {
        __index = function( t, k, v )
            local cap = query_time

            if type(k) == "number" then cap = cap + ( k / 1000 )
            else
                if not class.abilities[ k ] then k = "summon_demonic_tyrant" end
                cap = cap + action[ k ].cast
            end

            -- In SimC, k would be a numeric value to be interpreted but I don't see the point.
            -- We're only using it for SDT now, and I don't know what else we'd really use it for.

            -- So imps_spawned_during.summon_demonic_tyrant would be the syntax I'll use here.

            local n = 0

            for i, spawn in ipairs( guldan_v ) do
                if spawn > cap then break end
                if spawn > query_time then n = n + 1 end
            end

            return n
        end,
    } ) )


    -- Auras
    spec:RegisterAuras( {
        axe_toss = {
            id = 89766,
            duration = 4,
            max_stack = 1,
        },
        banish = {
            id = 710,
            duration = 30,
            max_stack = 1,
        },
        bile_spit = {
            id = 267997,
            duration = 10,
            max_stack = 1,
        },
        burning_rush = {
            id = 111400,
            duration = 3600,
            max_stack = 1,
        },
        corruption = {
            id = 146739,
            duration = 14,
            type = "Magic",
            max_stack = 1,
        },
        curse_of_tongues = {
            id = 1714,
            duration = 60,
            type = "Curse",
            max_stack = 1,
        },
        curse_of_weakness = {
            id = 702,
            duration = 120,
            type = "Curse",
            max_stack = 1,
        },
        dark_pact = {
            id = 108416,
            duration = 20,
            max_stack = 1,
        },
        demonic_calling = {
            id = 205146,
            duration = 20,
            type = "Magic",
            max_stack = 1,
        },
        demonic_circle = {
            id = 48018,
            duration = 900,
            max_stack = 1,
        },
        demonic_circle_teleport = {
            id = 48020,
        },
        demonic_core = {
            id = 264173,
            duration = 20,
            max_stack = 4,
        },
        demonic_power = {
            id = 265273,
            duration = 15,
            max_stack = 1,
            copy = "tyrant"
        },
        demonic_strength = {
            id = 267171,
            duration = 20,
            max_stack = 1,
        },
        doom = {
            id = 603,
            duration = function () return 20 * haste end,
            tick_time = function () return 20 * haste end,
            max_stack = 1,
        },
        drain_life = {
            id = 234153,
            duration = function () return 5 * haste * ( legendary.claw_of_endereth.enabled and 0.5 or 1 ) end,
            tick_time = function () return haste * ( legendary.claw_of_endereth.enabled and 0.5 or 1 ) end,
            max_stack = 1,
        },
        eye_of_kilrogg = {
            id = 126,
        },
        fear = {
            id = 118699,
            duration = 20,
            type = "Magic",
            max_stack = 1,
        },
        fel_domination = {
            id = 333889,
            duration = 15,
            type = "Magic",
            max_stack = 1,
        },
        felstorm = {
            id = 89751,
            duration = function () return 5 * haste end,
            tick_time = function () return 1 * haste end,
            max_stack = 1,

            generate = function ()
                local fs = buff.felstorm

                local name, _, count, _, duration, expires, caster = FindUnitBuffByID( "pet", 89751 )

                if name then
                    fs.count = 1
                    fs.applied = expires - duration
                    fs.expires = expires
                    fs.caster = "pet"
                    return
                end

                fs.count = 0
                fs.applied = 0
                fs.expires = 0
                fs.caster = "nobody"
            end,
        },
        from_the_shadows = {
            id = 270569,
            duration = 12,
            max_stack = 1,
        },
        howl_of_terror = {
            id = 5484,
            duration = 20,
            type = "Magic",
            max_stack = 1,
        },
        legion_strike = {
            id = 30213,
            duration = 6,
            max_stack = 1,
        },
        mortal_coil = {
            id = 6789,
            duration = 3,
            type = "Magic",
            max_stack = 1,
        },
        nether_portal = {
            id = 267218,
            duration = 15,
            max_stack = 1,
        },
        ritual_of_summoning = {
            id = 698,
        },
        shadowfury = {
            id = 30283,
            duration = 3,
            type = "Magic",
            max_stack = 1,
        },
        soul_leech = {
            id = 108366,
            duration = 15,
            max_stack = 1,
        },

        soul_link = {
            id = 108415,
        },
        soulstone = {
            id = 20707,
            duration = 900,
            max_stack = 1,
        },
        subjugate_demon = {
            id = 1098,
            duration = 300,
            max_stack = 1,
        },
        unending_breath = {
            id = 5697,
            duration = 600,
            max_stack = 1,
        },
        unending_resolve = {
            id = 104773,
            duration = 8,
            max_stack = 1,
        },

        dreadstalkers = {
            duration = 12,

            meta = {
                up = function ()
                    local exp = dreadstalkers_v[ #dreadstalkers_v ]
                    return exp and exp >= query_time or false
                end,
                down = function ( t ) return not t.up end,
                applied = function () local exp = dreadstalkers_v[ 1 ]; return exp and ( exp - 12 ) or 0 end,
                remains = function () local exp = dreadstalkers_v[ #dreadstalkers_v ]; return exp and max( 0, exp - query_time ) or 0 end,
                count = function ()
                    local c = 0
                    for i, exp in ipairs( dreadstalkers_v ) do
                        if exp >= query_time then c = c + 2 end
                    end
                    return c
                end,
            }
        },

        grimoire_felguard = {
            duration = 17,

            meta = {
                up = function () local exp = grim_felguard_v[ #grim_felguard_v ]; return exp and exp >= query_time or false end,
                down = function ( t ) return not t.up end,
                applied = function () local exp = grim_felguard_v[ 1 ]; return exp and ( exp - 12 ) or 0 end,
                remains = function () local exp = grim_felguard_v[ #grim_felguard_v ]; return exp and max( 0, exp - query_time ) or 0 end,
                count = function ()
                    local c = 0
                    for i, exp in ipairs( grim_felguard_v ) do
                        if exp > query_time then c = c + 1 end
                    end
                    return c
                end,
            }
        },

        vilefiend = {
            duration = 15,

            meta = {
                up = function () local exp = vilefiend_v[ #vilefiend_v ]; return exp and exp >= query_time or false end,
                down = function ( t ) return not t.up end,
                applied = function () local exp = vilefiend_v[ 1 ]; return exp and ( exp - 15 ) or 0 end,
                remains = function () local exp = vilefiend_v[ #vilefiend_v ]; return exp and max( 0, exp - query_time ) or 0 end,
                count = function ()
                    local c = 0
                    for i, exp in ipairs( vilefiend_v ) do
                        if exp > query_time then c = c + 1 end
                    end
                    return c
                end,
            }
        },

        wild_imps = {
            duration = 25,

            meta = {
                up = function () local exp = wild_imps_v[ #wild_imps_v ]; return exp and exp >= query_time or false end,
                down = function ( t ) return not t.up end,
                applied = function () local exp = wild_imps_v[ 1 ]; return exp and ( exp - 20 ) or 0 end,
                remains = function () local exp = wild_imps_v[ #wild_imps_v ]; return exp and max( 0, exp - query_time ) or 0 end,
                count = function ()
                    local c = 0
                    for i, exp in ipairs( wild_imps_v ) do
                        if exp > query_time then c = c + 1 end
                    end

                    -- Count queued HoG imps.
                    for i, spawn in ipairs( guldan_v ) do
                        if spawn <= query_time and ( spawn + 20 ) >= query_time then c = c + 1 end
                    end
                    return c
                end,
            }
        },

        other_demon = {
            duration = 20,

            meta = {
                up = function () local exp = other_demon_v[ 1 ]; return exp and exp >= query_time or false end,
                down = function ( t ) return not t.up end,
                applied = function () local exp = other_demon_v[ 1 ]; return exp and ( exp - 15 ) or 0 end,
                remains = function () local exp = other_demon_v[ 1 ]; return exp and max( 0, exp - query_time ) or 0 end,
                count = function ()
                    local c = 0
                    for i, exp in ipairs( other_demon_v ) do
                        if exp > query_time then c = c + 1 end
                    end
                    return c
                end,
            }
        },


        -- Azerite Powers
        forbidden_knowledge = {
            id = 279666,
            duration = 15,
            max_stack = 1,
        },
    } )


    local Glyphed = IsSpellKnownOrOverridesKnown

    -- Fel Imp          58959
    spec:RegisterPet( "imp",
        function() return Glyphed( 112866 ) and 58959 or 416 end,
        "summon_imp",
        3600 )

    -- Voidlord         58960
    spec:RegisterPet( "voidwalker",
        function() return Glyphed( 112867 ) and 58960 or 1860 end,
        "summon_voidwalker",
        3600 )

    -- Observer         58964
    spec:RegisterPet( "felhunter",
        function() return Glyphed( 112869 ) and 58964 or 417 end,
        "summon_felhunter",
        3600 )

    -- Fel Succubus     120526
    -- Shadow Succubus  120527
    -- Shivarra         58963
    spec:RegisterPet( "succubus",
        function()
            if Glyphed( 240263 ) then return 120526
            elseif Glyphed( 240266 ) then return 120527
            elseif Glyphed( 112868 ) then return 58963 end
            return 1863
        end,
        3600 )

    -- Wrathguard       58965
    spec:RegisterPet( "felguard",
        function() return Glyphed( 112870 ) and 58965 or 17252 end,
        "summon_felguard",
        3600 )

    spec:RegisterPet( "doomguard",
        11859,
        "ritual_of_doom",
        300 )
    
    
    --[[ Demonic Tyrant
    spec:RegisterPet( "demonic_tyrant",
        135002,
        "summon_demonic_tyrant",
        15 ) ]]
    
    spec:RegisterTotem( "demonic_tyrant", 135002 )    
    spec:RegisterTotem( "vilefiend", 1616211 )
    spec:RegisterTotem( "grimoire_felguard", 136216 )
    spec:RegisterTotem( "dreadstalker", 1378282 )


    spec:RegisterStateExpr( "extra_shards", function () return 0 end )

    --[[ spec:RegisterVariable( "tyrant_ready", function ()
        if cooldown.summon_demonic_tyrant.remains > 5 then return false end
        if talent.demonic_strength.enabled and not talent.demonic_consumption.enabled and cooldown.demonic_strength.ready then return false end
        if talent.nether_portal.enabled and cooldown.nether_portal.ready then return false end
        if talent.grimoire_felguard.enabled and cooldown.grimoire_felguard.ready then return false end
        if talent.summon_vilefiend.enabled and cooldown.summon_vilefiend.ready then return false end
        if cooldown.call_dreadstalkers.ready then return false end
        if buff.demonic_core.up and soul_shard < 4 and ( talent.demonic_consumption.enabled or buff.nether_portal.down ) then return false end
        if soul_shard < ( buff.nether_portal.up and 1 or 5 ) then return false end
        return true
    end ) ]]


    -- Abilities
    spec:RegisterAbilities( {
        axe_toss = {
            id = 119914,
            known = function () return IsSpellKnownOrOverridesKnown( 119914 ) end,
            cast = 0,
            cooldown = 30,
            gcd = "spell",

            startsCombat = true,

            toggle = "interrupts",
            interrupt = true,

            debuff = "casting",
            readyTime = state.timeToInterrupt,

            usable = function () return pet.exists end,
            handler = function ()
                interrupt()
                applyDebuff( "target", "axe_toss", 4 )
            end,

            copy = 119914
        },


        banish = {
            id = 710,
            cast = 1.5,
            cooldown = 0,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
                applyDebuff( "target", "banish", 30 )
            end,
        },


        bilescourge_bombers = {
            id = 267211,
            cast = 0,
            cooldown = 30,
            gcd = "spell",

            spend = 2,
            spendType = "soul_shards",

            talent = "bilescourge_bombers",

            startsCombat = true,

            handler = function ()
            end,
        },


        burning_rush = {
            id = 111400,
            cast = 0,
            cooldown = 0,
            gcd = "spell",

            startsCombat = true,

            talent = "burning_rush",

            handler = function ()
                if buff.burning_rush.up then removeBuff( "burning_rush" )
                else applyBuff( "burning_rush", 3600 ) end
            end,
        },


        call_felhunter = {
            id = 212619,
            cast = 0,
            cooldown = 24,
            gcd = "spell",

            spend = 0.01,
            spendType = "mana",

            startsCombat = true,
            texture = 136174,

            toggle = "interrupts",
            interrupt = true,

            pvptalent = "call_felhunter",

            debuff = "casting",
            readyTime = state.timeToInterrupt,

            handler = function ()
                interrupt()
            end,
        },


        -- PvP:master_summoner.
        call_dreadstalkers = {
            id = 104316,
            cast = function () if pvptalent.master_summoner.enabled then return 0 end
                return buff.demonic_calling.up and 0 or ( ( level > 53 and 1.5 or 2 ) * haste )
            end,
            cooldown = 20,
            gcd = "spell",

            spend = function () return buff.demonic_calling.up and 0 or 2 end,
            spendType = "soul_shards",

            startsCombat = true,

            handler = function ()
                summon_demon( "dreadstalkers", 12, 2 )
                summonPet( "dreadstalker", 12 )
                removeStack( "demonic_calling" )

                if talent.from_the_shadows.enabled then applyDebuff( "target", "from_the_shadows" ) end
            end,
        },


        --[[ command_demon = {
            id = 119898,
            cast = 0,
            cooldown = 0,
            gcd = "spell",

            startsCombat = true,

            handler = function ()
                if pet.felguard.up then runHandler( "axe_toss" )
                elseif pet.felhunter.up then runHandler( "spell_lock" )
                elseif pet.voidwalker.up then runHandler( "shadow_bulwark" )
                elseif pet.succubus.up then runHandler( "seduction" )
                elseif pet.imp.up then runHandler( "singe_magic" ) end
            end,
        }, ]]


        create_healthstone = {
            id = 6201,
            cast = function () return 3 * haste end,
            cooldown = 0,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
            end,
        },


        create_soulwell = {
            id = 29893,
            cast = function () return 3 * haste end,
            cooldown = 120,
            gcd = "spell",

            spend = 0.05,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
            end,
        },


        dark_pact = {
            id = 108416,
            cast = 0,
            cooldown = 60,
            gcd = "spell",

            startsCombat = true,

            talent = "dark_pact",

            handler = function ()
                applyBuff( "dark_pact", 20 )
            end,
        },


        demonbolt = {
            id = 264178,
            cast = function () return ( buff.demonic_core.up and 0 or 4.5 ) * haste end,
            cooldown = 0,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
                if buff.forbidden_knowledge.up and buff.demonic_core.down then
                    removeBuff( "forbidden_knowledge" )
                end

                removeStack( "demonic_core" )
                removeStack( "decimating_bolt" )
                gain( 2, "soul_shards" )
            end,
        },


        demonic_circle = {
            id = 48018,
            cast = 0.5,
            cooldown = 10,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = false,
            nobuff = "demonic_circle",

            handler = function ()
                applyBuff( "demonic_circle" )
            end,
        },


        demonic_circle_teleport = {
            id = 48020,
            cast = 0,
            cooldown = 30,
            gcd = "spell",

            spend = 0.03,
            spendType = "mana",

            startsCombat = false,

            talent = "demonic_circle",
            buff = "demonic_circle",

            handler = function ()
                if conduit.demonic_momentum.enabled then applyBuff( "demonic_momentum" ) end
            end,

            auras = {
                -- Conduit
                demonic_momentum = {
                    id = 339412,
                    duration = 5,
                    max_stack = 1
                }
            }
        },


        demonic_gateway = {
            id = 111771,
            cast = function () return legendary.pillars_of_the_dark_portal.enabled and 0 or 2 end,
            cooldown = 10,
            gcd = "spell",

            spend = 0.03,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
            end,
        },


        demonic_strength = {
            id = 267171,
            cast = 0,
            cooldown = 60,
            gcd = "spell",

            startsCombat = false,
            nobuff = "felstorm",

            handler = function ()
                applyBuff( "demonic_strength" )
            end,
        },


        doom = {
            id = 603,
            cast = 0,
            cooldown = 0,
            gcd = "spell",

            spend = 0.01,
            spendType = "mana",

            startsCombat = true,

            talent = "doom",

            cycle = "doom",
            min_ttd = function () return 3 + debuff.doom.duration end,

            -- readyTime = function () return IsCycling() and 0 or debuff.doom.remains end,
            -- usable = function () return IsCycling() or ( target.time_to_die < 3600 and target.time_to_die > debuff.doom.duration ) end,
            handler = function ()
                applyDebuff( "target", "doom" )
            end,
        },


        drain_life = {
            id = 234153,
            cast = function () return 5 * haste * ( legendary.claw_of_endereth.enabled and 0.5 or 1 ) end,
            cooldown = 0,
            channeled = true,
            gcd = "spell",

            spend = function () return debuff.soul_rot.up and 0 or 0.03 end,
            spendType = "mana",

            startsCombat = true,

            start = function ()
                applyDebuff( "drain_life" )
            end,

            finish = function ()
                if conduit.accrued_vitality.enabled then applyBuff( "accrued_vitality" ) end
            end,
        },


        eye_of_kilrogg = {
            id = 126,
            cast = function () return 2 * haste end,
            cooldown = 0,
            gcd = "spell",

            spend = 0.03,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
            end,
        },


        fear = {
            id = 5782,
            cast = function () return 1.7 * haste end,
            cooldown = 0,
            gcd = "spell",

            spend = 0.05,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
                applyDebuff( "target", "fear" )
            end,
        },


        fel_domination = {
            id = 333889,
            cast = 0,
            cooldown = function () return 180 + conduit.fel_celerity.mod * 0.001 end,
            gcd = "spell",

            startsCombat = false,
            texture = 237564,

            essential = true,
            nomounted = true,
            nobuff = "grimoire_of_sacrifice",

            handler = function ()
                applyBuff( "fel_domination" )
            end,
        },


        grimoire_felguard = {
            id = 111898,
            cast = 0,
            cooldown = 120,
            gcd = "spell",

            spend = 1,
            spendType = "soul_shards",

            toggle = "cooldowns",

            startsCombat = true,

            handler = function ()
                summon_demon( "grimoire_felguard", 15 )
                applyBuff( "grimoire_felguard" )
                summonPet( "grimoire_felguard" )
            end,
        },


        hand_of_guldan = {
            id = 105174,
            cast = function () return 1.5 * haste end,
            cooldown = 0,
            gcd = "spell",

            spend = 1,
            spendType = "soul_shards",

            startsCombat = true,

            -- usable = function () return soul_shards.current >= 3 end,
            handler = function ()
                extra_shards = min( 2, soul_shards.current )
                if Hekili.ActiveDebug then Hekili:Debug( "Extra Shards: %d", extra_shards ) end
                spend( extra_shards, "soul_shards" )
                update_tyrant_readiness( 1 + extra_shards )
                insert( guldan_v, query_time + 0.6 )
                if extra_shards > 0 then insert( guldan_v, query_time + 0.8 ) end
                if extra_shards > 1 then insert( guldan_v, query_time + 1 ) end
            end,
        },


        health_funnel = {
            id = 755,
            cast = function () return 5 * haste end,
            cooldown = 0,
            gcd = "spell",

            channeled = true,

            startsCombat = false,
            texture = 607852,

            usable = function () return pet.alive and pet.health_pct < 100, "requires injured demon" end,

            start = function ()
                applyBuff( "health_funnel" )
            end,
        },


        implosion = {
            id = 196277,
            cast = 0,
            cooldown = 0,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = true,
            texture = 2065588,

            velocity = 30,

            usable = function ()
                if buff.wild_imps.stack < 3 and azerite.explosive_potential.enabled then return false, "too few imps for explosive_potential"
                elseif buff.wild_imps.stack < 1 then return false, "no imps available" end
                return true
            end,

            handler = function ()
                if azerite.explosive_potential.enabled and buff.wild_imps.stack >= 3 then applyBuff( "explosive_potential" ) end
                if legendary.implosive_potential.enabled and active_enemies > 2 then
                    if buff.implosive_potential.down then stat.haste = stat.haste + 0.05 * buff.wild_imps.stack end
                    applyBuff( "implosive_potential" )
                end
                consume_demons( "wild_imps", "all" )
            end,

            auras = {
                implosive_potential = {
                    id = 337139,
                    duration = 8,
                    max_stack = 1
                }
            }
        },


        mortal_coil = {
            id = 6789,
            cast = 0,
            cooldown = 45,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = true,
            texture = 607853,

            handler = function ()
                applyDebuff( "target", "mortal_coil" )
            end,
        },


        nether_portal = {
            id = 267217,
            cast = function () return 2.5 * haste end,
            cooldown = 180,
            gcd = "spell",

            spend = 1,
            spendType = "soul_shards",

            toggle = "cooldowns",

            startsCombat = false,
            texture = 2065615,

            handler = function ()
                applyBuff( "nether_portal" )
            end,
        },


        power_siphon = {
            id = 264130,
            cast = 0,
            cooldown = 30,
            gcd = "spell",

            startsCombat = false,
            texture = 236290,

            talent = "power_siphon",

            readyTime = function ()
                if buff.wild_imps.stack >= 2 then return 0 end

                local imp_deficit = 2 - buff.wild_imps.stack

                for i, imp in ipairs( guldan_v ) do
                    if imp > query_time then
                        imp_deficit = imp_deficit - 1
                        if imp_deficit == 0 then return imp - query_time end
                    end
                end

                return 3600
            end,

            handler = function ()
                local num = min( 2, buff.wild_imps.count )
                consume_demons( "wild_imps", num )

                addStack( "demonic_core", 20, num )
            end,
        },


        ritual_of_summoning = {
            id = 698,
            cast = 0,
            cooldown = 120,
            gcd = "spell",

            spend = 0,
            spendType = "mana",

            startsCombat = true,
            texture = 136223,

            handler = function ()
            end,
        },


        shadow_bolt = {
            id = 686,
            cast = function () return 2 * haste end,
            cooldown = 0,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = true,
            texture = 136197,

            handler = function ()
                gain( 1, "soul_shards" )

                if legendary.balespiders_burning_core.enabled then
                    addStack( "balespiders_burning_core", nil, 1 )
                end
            end,

            auras = {
                balespiders_burning_core = {
                    id = 337161,
                    duration = 15,
                    max_stack = 4
                }
            }
        },


        shadowfury = {
            id = 30283,
            cast = function () return 1.5 * haste end,
            cooldown = function () return talent.darkfury.enabled and 45 or 60 end,
            gcd = "spell",

            startsCombat = true,
            texture = 607865,

            handler = function ()
                applyDebuff( "target", "shadowfury" )
            end,
        },


        soul_strike = {
            id = 264057,
            cast = 0,
            cooldown = 10,
            gcd = "spell",

            startsCombat = true,
            texture = 1452864,

            usable = function () return pet.felguard.up and pet.alive, "requires living felguard" end,
            handler = function ()
                gain( 1, "soul_shards" )
            end,
        },


        soulstone = {
            id = 20707,
            cast = 3,
            cooldown = 600,
            gcd = "spell",

            spend = 0.01,
            spendType = "mana",

            toggle = "cooldowns",

            startsCombat = false,
            texture = 136210,

            handler = function ()
                applyBuff( "soulstone" )
            end,
        },


        subjugate_demon = {
            id = 1098,
            cast = 3,
            cooldown = 0,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = false,
            texture = 136154,

            usable = function () return target.is_demon and target.level < level + 2, "requires demon enemy" end,
            handler = function ()
                summonPet( "subjugate_demon" )
            end,
        },


        summon_demonic_tyrant = {
            id = 265187,
            cast = function () return 2 * haste end,
            cooldown = function () return ( essence.vision_of_perfection.enabled and 0.87 or 1 ) * 90 end,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            toggle = "cooldowns",

            startsCombat = true,
            texture = 2065628,

            handler = function ()
                summonPet( "demonic_tyrant", 15 )
                summon_demon( "demonic_tyrant", 15 )

                applyBuff( "demonic_power", 15 )

                tyrant_ready = false

                --[[ if talent.demonic_consumption.enabled then
                    consume_demons( "wild_imps", "all" )
                end ]]

                extend_demons()

                if azerite.baleful_invocation.enabled or level > 57 then gain( 5, "soul_shards" ) end
            end,

            auras = {
                -- Conduit
                -- Note:  Should set up a queued event for this to start when Tyrant finishes.
                tyrants_soul = {
                    id = 339766,
                    duration = 15,
                    max_stack = 1
                }
            }
        },


        summon_felguard = {
            id = 30146,
            cast = function () return ( buff.fel_domination.up and 0.5 or 6 ) * haste end,
            cooldown = 0,
            gcd = "spell",

            spend = function () return buff.fel_domination.up and 0 or 1 end,
            spendType = "soul_shards",

            startsCombat = false,
            essential = true,

            bind = "summon_pet",
            nomounted = true,

            usable = function () return not pet.exists, "cannot have an existing pet" end,
            handler = function ()
                removeBuff( "fel_domination" )
                summonPet( "felguard", 3600 )
            end,

            copy = { "summon_pet", 112870 }
        },


        summon_vilefiend = {
            id = 264119,
            cast = function () return 2 * haste end,
            cooldown = 45,
            gcd = "spell",

            spend = 1,
            spendType = "soul_shards",

            startsCombat = true,
            texture = 1616211,

            handler = function ()
                summon_demon( "vilefiend", 15 )
                summonPet( "vilefiend", 15 )
            end,
        },


        unending_breath = {
            id = 5697,
            cast = 0,
            cooldown = 0,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            startsCombat = true,

            handler = function ()
            end,
        },


        unending_resolve = {
            id = 104773,
            cast = 0,
            cooldown = 180,
            gcd = "spell",

            spend = 0.02,
            spendType = "mana",

            toggle = "defensives",

            startsCombat = true,

            handler = function ()
            end,
        },


        --[[ wartime_ability = {
            id = 264739,
            cast = 0,
            cooldown = 0,
            gcd = "spell",

            startsCombat = true,

            handler = function ()
            end,
        }, ]]
    } )


    spec:RegisterOptions( {
        enabled = true,

        aoe = 3,

        nameplates = false,
        nameplateRange = 8,

        cycle = true,

        damage = true,
        damageExpiration = 6,

        potion = "spectral_intellect",

        package = "Demonology",
    } )


    spec:RegisterPack( "Demonology", 20210823, [[dW0SLbqisbEevrCjLKQ2KsLpPKYOiLCksrRsvc8ksPMfvPUfvrTlr9lrkdtKkhdP0YOk5zkjzAufPRrk02uskFJuqnosb5CQsswhsrnpKc3JQQ9PK4GIuvwOsQEisrMOQe6IIuL2OQeu(OQKuJuKQqNuKQQvQuAMQsQDkI8tvjrdvKQOLQkb5PizQuf2QQeu9vvjHXQkr7Lk)LKbd5WkwmP6XOmzuDzWMvIplsgnvLtl51QsnBkUnL2TWVLA4QIJRKuz5iEoutN46QQTlcFxPy8kvDEKQ1lsvW8frTFv2rRZdhfFeWLKxPZlAtNgYRvLtNgAvAmDRMJsO)aoQNH9EsboQySGJ6fbBhTPtr3r9m0n9WDE4OW9NWahfvzPjhL(Vms6pC6ok(iGljVsNx0MonKxRkNon0Q0y60OJc)amxsETARMJYxX5q40DuCaZCuViy7OnDk6h6vmetZEFB9jYdMMtlTuL47RNzTnnCz)MrQoyKzrsdxwwA3203p1hlhYRv59H8kDEr7T92st(MifGP5BRNpe1dymh61n7D(265d9kdd9drawBTqWp0lc2o0BJCOhc4zwB1h5q1YHk5qf(qvGLjKdPvtoKVHWzdwo0stoKEJXawZ8T1Zhk9S3aKdrvp(64qJX0Ba(HEiGNzTvFKdj9HEin7qvGLjKd9IGTd92i5BRNpu6zI0ZdjJbc5qviaH8FK8T1Zhk9LOl(HOw3ZRKESF1hc)m2dTXhehIE)xJahkA5qJE)Ldj9HWFRTJdnhYd6KjK8T1Zh6fMbW(yKzrs7fEBgPmWHOAtciKdXMGbgvTCiMVjsb8dj9HQqac5)iQAj7OmfwWopCuytVrjKkEdc25HljADE4OGy0na3TUJAys1HJc3FJbePIukYxNUJIrkbi14OyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8drJdjdjfizEHLjyWHs7qA8q7oKuw4qRCOedPgDdKxkcwucDYeIsklCipFiToKmKuGK5fwMGbhkTdPXdPPJkgl4OW93yarQiLI81P7exsE58WrbXOBaUBDh1WKQdhf(h6MU5QXcIp6yXrXiLaKACuSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pata(HOXHKHKcKmVWYem4qPDinEODhsklCOvouIHuJUbYlfblkHozcrjLfoKNpKwhsgskqY8cltWGdL2H04H00rfJfCu4FOB6MRgli(OJfN4sAvopCuqm6gG7w3rXbmJups1HJ6vs4Xem4q(g8HMdrRxhcdSo4hIdMH(HMGFOcFiXhqGLMahc)UEEa(HwAYHwkcwoKh0jtihs6dzQao0)5qBkX3HeFWHiawCuXybhfyFOtGXOAcpMGbokgPeGuJJI1TH3BIm(BTDO4GTJIPe6KjKmbStfyfS)bycWpenoKwhsgskqY8cltWGdL2H04H08qAFiA96q7oeRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGFOvoKwhsRdP1HKHKcKmVWYem4qPDinEinpK2hIwVoKMhYZhIwnEinp0UdjLfo0khkXqQr3a5LIGfLqNmHOKYchYZhsRdP1HKHKcKmVWYem4qPDinEinpK2hIwVoKMoQHjvhokW(qNaJr1eEmbdCItCu(EucPI3yNhUKO15HJcIr3aC36oQySGJcxXY3OszgEnstWkWQBaRJAys1HJcxXY3OszgEnstWkWQBaRtCj5LZdhfeJUb4U1DuXybhfUILVrn4NImHGvGv3awh1WKQdhfUILVrn4NImHGvGv3awN4ehfRtaXeIA0ltj0DE4sIwNhokigDdWDR7OyKsasnokC)n6vWZPiDcqvrIkvtgP6idXOBa(H2DiToeRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGFiACiVs3Hso5dX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4hALdTQ0DinDudtQoCu4(BuKwCIljVCE4OGy0na3TUJIrkbi14OW93OxbpVuGHR6fLUPX42IZqm6gGFODh6bKmhSDumLqNmHKhMujah1WKQdhfU)gfPfN4sAvopCuqm6gG7w3rXiLaKACu4(B0RGN3ugUY3peLmmPy4meJUb4oQHjvhokC)nksloXLKN68WrbXOBaUBDhfJucqQXrP1HW93OxbpBGHR0PRG9J9XazigDdWpuYjFiC)n6vWZVHevGvDNEamvKkdXOBa(H08q7oKwh6bKmhSDumLqNmHKhMujGdT7q4(BuyFdHFiACiVouYjFin4qpGK5GTJIPe6KjK8WKkbCODhI1TH3BIm(BTDO4GTJIPe6KjKmbStfyfS)bycWp0khYtt3H00rnmP6WrXbwzhPIuk92ioXLKgDE4OGy0na3TUJIrkbi14O06q4(B0RGNxAskqVjbOiqcGuaodXOBa(Hso5dP1HW93OxbpNOnJugqHBtciKmeJUb4hA3H0GdH7VrVcE(nKOcSQ70dGPIuzigDdWpKMhsZdT7qAWHEajZbBhftj0jti5HjvcWrnmP6WrXbwzhPIuk92ioXL0Q58WrbXOBaUBDh1WKQdh1IbW(yKzrCumsjaPghfU)g9k45eTzKYakCBsaHKHy0na3rvHaeY)ru1IJs)VSKt0MrkdOWTjbes()4exsAyNhokigDdWDR7OyKsasnokC)n6vWZS2QpIYc8sgP6idXOBa(H2DOhqYCW2rXucDYesEysLaCudtQoCuyw)jvKsjL4dCIljnKZdhfeJUb4U1DumsjaPghLgCiC)n6vWZS2QpIYc8sgP6idXOBaUJAys1HJcZ6pPIukPeFGtCj9QCE4OGy0na3TUJIrkbi14OEajZbBhftj0jti5Hjvc4q7oeU)gf23q4hY)HsNJAys1HJQSpqWRiLInYGfs)4dCItCucDYeIcdY)X5HljADE4OGy0na3TUJIrkbi14OyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8drJdrRgDudtQoCubi(aI6PjYyCIljVCE4OGy0na3TUJIrkbi14OyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8drJdrRg(qE(qADOHjvhz83A7qXbBhftj0jtizypW(cOKYchs7dnmP6iJ9n8EJsVnsg2dSVakPSWH08q7oKwhI1TH3BImBmgfNadhlJ5nqWzcyNkWhIghIwn8H88H06qdtQoY4V12HId2okMsOtMqYWEG9fqjLfoK2hAys1rg)T2oujkdSuqWZWEG9fqjLfoK2hAys1rg7B49gLEBKmShyFbuszHdP5Hso5d9asMtGHJLX8gizcyNkWhALdX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4hs7dnmP6iJ)wBhkoy7OykHozcjd7b2xaLuw4qA6OgMuD4OsrkBxeqTaMu)HWDIlPv58WrbXOBaUBDhfJucqQXrP1HyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8drJdrRgpKNpKwhAys1rg)T2ouCW2rXucDYesg2dSVakPSWH08q7oKwhI1TH3BImBmgfNadhlJ5nqWzcyNkWhIghIwnEipFiTo0WKQJm(BTDO4GTJIPe6KjKmShyFbuszHdP9HgMuDKXFRTdvIYalfe8mShyFbuszHdP5Hso5d9asMtGHJLX8gizcyNkWhALdX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4hs7dnmP6iJ)wBhkoy7OykHozcjd7b2xaLuw4qAEinpuYjFiToKgCiYpGLMKcYBkZcb4yfUsvgvVOW)hGunrH)wBhvKkdXOBa(H2Diw3gEVjY4V12HId2okMsOtMqYeWovGvW(hGja)qRCipnDhsth1WKQdhf(BTDOsugyPGG7exsEQZdhfeJUb4U1DumsjaPghfRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGFiACiA96qE(qADOHjvhz83A7qXbBhftj0jtizypW(cOKYchs7dnmP6iJ9n8EJsVnsg2dSVakPSWH08q7oKuw4qRCOedPgDdKxkcwucDYeIsklCipFiA96qE(qdtQoYSXyuCcmCSmM3abNH9a7lGsklCiTp0WKQJm(BTDO4GTJIPe6KjKmShyFbuszHdP9HgMuDKX(gEVrP3gjd7b2xaLuwWrnmP6WrXgJrXjWWXYyEdeStCjPrNhokigDdWDR7OyKsasnokPSWHw5qjgsn6giVueSOe6KjeLuw4q7oKwh6bKmNadhlJ5nqYdtQeWH2DOhqYCcmCSmM3ajta7ub(qRCOHjvhz83A7qXbBhftj0jtizypW(cOKYchsZdT7qADin4qYyGqY4V12HkrzGLccEgIr3a8dLCYh6bKCIYalfe88WKkbCinp0UdP1HW93OW(gc)q(pu6ouYjFiTo0dizobgowgZBGKhMujGdT7qpGK5ey4yzmVbsMa2Pc8HOXHgMuDKXFRTdfhSDumLqNmHKH9a7lGsklCiTp0WKQJm23W7nk92izypW(cOKYchsZdLCYhsRd9asorzGLccEEysLao0Ud9asorzGLccEMa2Pc8HOXHgMuDKXFRTdfhSDumLqNmHKH9a7lGsklCiTp0WKQJm23W7nk92izypW(cOKYchsZdLCYhsRdP)xwYPiLTlcOwatQ)q45)ZH2Di9)YsofPSDra1cys9hcpta7ub(q04qdtQoY4V12HId2okMsOtMqYWEG9fqjLfoK2hAys1rg7B49gLEBKmShyFbuszHdP5H00rnmP6WrH)wBhkoy7OykHozcXjoXrXHL5BeNhUKO15HJcIr3aC36okoGzK6rQoCuP39a7la)qqcGq)qszHdj(GdnmPjhQWhAsmLz0nq2rnmP6WrHFaJrzA2BN4sYlNhoQHjvhok2ymQfW47hcqCuqm6gG7w3jUKwLZdh1WKQdh1ShusJXokigDdWDR7exsEQZdh1WKQdhfhs0FIYoPkMJcIr3aC36oXLKgDE4OGy0na3TUJAys1HJIngJAys1HYuyXrzkSOIXcokHuXBqWoXL0Q58WrbXOBaUBDhfJucqQXrrGfcG9n6gWrnmP6WrX726exsAyNhokigDdWDR7OyKsasnokC)n6vWZPiDcqvrIkvtgP6idXOBa(Hso5dH7VrVcEEPadx1lkDtJXTfNHy0na)qjN8HW93OxbpZAR(iklWlzKQJmeJUb4hk5KpeRtaXesoagPnnH7OWcPyIljADudtQoCuSXyudtQouMcloktHfvmwWrX6eqmHOg9YucDN4ssd58WrbXOBaUBDh1WKQdhfBmg1WKQdLPWIJYuyrfJfCucDYeIcdY)XjUKEvopCuqm6gG7w3rXiLaKACuADiw3gEVjY4V12HId2okMsOtMqYeWovGvW(hGja)q04q0MUdT7qszHdTYHsmKA0nqEPiyrj0jtikPSWH88HOnDhk5KpeU)g9k4zcSubWvpJzeidXOBa(H2Diw3gEVjY4V12HId2okMsOtMqYeWovGvW(hGja)q04qRsdDinDudtQoCupTuD4exs0MoNhokigDdWDR7OyKsasnoQhqYCW2rXucDYesEysLaCuyHumXLeToQHjvhok2ymQHjvhktHfhLPWIkgl4O6umUtCjrlTopCuqm6gG7w3rXiLaKACuADin4qKFalnjfK3uMfcWXkCLQmQErH)paPAIc)T2oQivgIr3a8dT7qSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pata(Hw5qVQdP5Hso5dP1HEajZbBhftj0jti5Hjvc4q7o0dizoy7OykHozcjta7ub(q04qR2HEbhkfJNTZ(dPPJAys1HJId2okMcleisj(CIljA9Y5HJcIr3aC36okgPeGuJJI1TH3BIm(BTDO4GTJIPe6KjKmbStfyfS)bycWp0khYR0DipFinEOxWH0Gdr(bS0KuqEtzwiahRWvQYO6ff()aKQjk83A7OIuzigDdWDudtQoCuSXyuCcmCSmM3ab7exs0UkNhokigDdWDR7OyKsasnok9)YsEtz4QY(GZyzyVp0khI2dT7q6)LLmhSDumfRjqgld79HOXHwLJAys1HJ6P3aefUE81HtCjrRN68WrbXOBaUBDhfJucqQXrP)xwYcDYesM3BIdT7qSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pata(Hw5qA0rnmP6WrPxgaZ6pjfO0BRoqWoXLeTA05HJcIr3aC36okgPeGuJJAysLauqa2cWhALdr7H0(qADiAp0l4qYyGqY4HrQLIbCfU)gCgIr3a8dP5H2Di9)YsEtz4QY(GZyzyVp0k(p0QDODhs)VSKf6KjKmV3ehA3HyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8dTYH0OJAys1HJQSpMgxD4exs0UAopCuqm6gG7w3rXiLaKACudtQeGccWwa(qRCiVo0UdP)xwYBkdxv2hCgld79HwX)HwTdT7q6)LLSqNmHK59M4q7oeRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGFOvoKgp0UdPbhI8dyPjPGCzFmnUsaQNwGqQXKHy0na)q7oKwhsdoKmgiK8cPTkXhOW(gEVbNHy0na)qjN8H0)ll5fsBvIpqH9n8Edo)FoKMoQHjvhoQY(yAC1HtCjrRg25HJcIr3aC36okgPeGuJJAysLauqa2cWhALd51H2Di9)YsEtz4QY(GZyzyVp0k(p0QDODhs)VSKl7JPXvcq90cesnMmbStf4drJd51H2DiYpGLMKcYL9X04kbOEAbcPgtgIr3aCh1WKQdhvzFmnU6WjUKOvd58WrbXOBaUBDhfJucqQXrP)xwYBkdxv2hCgld79HwX)HO1RdT7qYyGqY4(BuSo4)sYqm6gGFODhsgdesEH0wL4duyFdV3GZqm6gGFODhI8dyPjPGCzFmnUsaQNwGqQXKHy0na)q7oK(Fzjl0jtizEVjo0UdX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4hALdPrh1WKQdhvzFmnU6WjUKO9v58WrbXOBaUBDhfJucqQXrP3y8H2DiPSGsAfVGdrJdTQ05OgMuD4OsrkBxeqTaMu)HWDIljVsNZdhfeJUb4U1DumsjaPghLEJXhA3HKYckPv8coenoKxAih1WKQdhf(BTDOsugyPGG7exsErRZdhfeJUb4U1DumsjaPghLEJXhA3HKYckPv8coenoeTA0rnmP6WrH)wBhkoy7OykHozcXjUK8YlNhokigDdWDR7OyKsasnokC)nkSVHWpK)dPrh1WKQdhLVj4QErL6B4t4exsETkNhokigDdWDR7OgMuD4O8nbx1lQuFdFchfhWms9ivhoQ0)YHErcmCSmM3abFOHahAmey40p0WKkb49HI(qba8dj9HWtc4qyFdHJDumsjaPghfU)gf23q4hAf)hAvhA3H06qpGK5ey4yzmVbsEysLaouYjFOhqYCW2rXucDYesEysLaoKMoXLKxEQZdhfeJUb4U1DumsjaPghfU)gf23q4hAf)hI2dT7q6)LLCaIpGOEAImM8)5q7oeRBdV3ez2ymkobgowgZBGGZeWovGp0khYRd9coukgpBN9oQHjvhokFtWv9Ik13WNWjUK8sJopCuqm6gG7w3rXiLaKACu4(BuyFdHFOv8FiAp0UdX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4hIghkfJNTZ(dT7qszHdTYHO1Rd55dLIXZ2z)H2DiToK(FzjZjWWXYyEdeC()CODhs)VSK5ey4yzmVbcota7ub(qRCOHjvhzFtWv9Ik13WNid7b2xaLuw4qAFOHjvhz83A7qXbBhftj0jtizypW(cOKYchsZdT7qADin4qYyGqY4V12HkrzGLccEgIr3a8dLCYhs)VSKtugyPGGN)phsth1WKQdhLVj4QErL6B4t4exsETAopCuqm6gG7w3rXiLaKACuAWHyDciMqYjGq8rN4OWcPyIljADudtQoCuSXyudtQouMcloktHfvmwWrX6eqmHOg9YucDN4sYlnSZdhfeJUb4U1DudtQoCu4(BuyHuVbhfhWms9ivhoQxrj(6VCiQHrQLIb8dr1Fd27dr1FZHOes9gouHpewiDKcihs8nXHErW2HEBeVpeUpujhY3Gp0CiFvkFa5qpKQjLq3rXiLaKACuAWHKXaHKXdJulfd4kC)n4meJUb4oXLKxAiNhokigDdWDR7OgMuD4O4GTd92iokoGzK6rQoCuupqWp0lc2ok2HOPMa4dT0Kdr1FZHO8neo(q)qkZH8Gozc5qSUn8EtCOcFiMPXWHK(qey40DumsjaPghL(FzjZbBhftXAcKjWWKdT7q4(BuyFdHFiACip9q7oeRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGFOvoKxPZjUK86v58WrbXOBaUBDh1WKQdhfhSDO3gXrXbmJups1HJ6f)KksDipOtMqoegK)J3hc)ab)qViy7OyhIMAcGp0stoev)nhIY3q4yhfJucqQXrP)xwYCW2rXuSMazcmm5q7oeU)gf23q4hIghYtp0UdX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4hIghIwVCIlPvLoNhokigDdWDR7OyKsasnok9)YsMd2okMI1eitGHjhA3HW93OW(gc)q04qE6H2DiToK(FzjZbBhftXAcKXYWEFOvoKxhk5KpKmgiKmEyKAPyaxH7VbNHy0na)qA6OgMuD4O4GTd92ioXL0QO15HJcIr3aC36okgPeGuJJs)VSK5GTJIPynbYeyyYH2DiC)nkSVHWpenoKNEODhAysLauqa2cWhALdrRJAys1HJId2o0BJ4exsRYlNhoQHjvhokC)nkSqQ3GJcIr3aC36oXL0QwLZdhfeJUb4U1DudtQoCuSXyudtQouMcloktHfvmwWrX6eqmHOg9YucDN4sAvEQZdhfeJUb4U1DudtQoCu(MGR6fvQVHpHJIdygPEKQdhv6F5q07)HytCOuGCi9H9(qsFinEiQ(BoeLVHWXhshwAcCOxKadhlJ5nqWhI1TH3BIdv4drGHt37dvYA4d1Vh6hs6dHFGGFiXhypu0BCumsjaPghfU)gf23q4hAf)hAvhA3HyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8dTYH8sJhA3H06qYyGqYCW2rXuSXyQivgIr3a8dLCYhI1TH3BImBmgfNadhlJ5nqWzcyNkWhALdP1H06qA8qE(q4(BuyFdHFinp0l4qdtQoYyFdV3O0BJKH9a7lGsklCinpK2hAys1r23eCvVOs9n8jYWEG9fqjLfoKMoXL0Q0OZdhfeJUb4U1DudtQoCu8UTokgPeGuJJIalea7B0nWH2DiPSWHw5qjgsn6giVueSOe6KjeLuwWrXOZmGsgskqWUKO1jUKw1Q58WrnmP6WrH9n8EJsVnIJcIr3aC36oXjoQhcWAR(iopCjrRZdhfeJUb4U1DudtQoCulGrXBBfJuD4O4aMrQhP6WrLE3dSVa8dPdlnboeRT6JCiDivf48HsFmg8i4dfD4zFdXU8nhAys1b(qDyONDumsjaPghLuw4qRCO0DODhsdo0di5XujaN4sYlNhoQHjvhok83A7qTaMu)HWDuqm6gG7w3jUKwLZdhfeJUb4U1DuXybhL0wq1lkBhyH0FSI1bwiFMuDGDudtQoCusBbvVOSDGfs)XkwhyH8zs1b2jUK8uNhokigDdWDR7OIXcokCBGXhwHbgbeLamFrT6(GJAys1HJc3gy8HvyGrarjaZxuRUp4exsA05HJAys1HJAXayFmYSiokigDdWDR7exsRMZdhfeJUb4U1DumsjaPghL(FzjVPmCvzFWzSmS3hALdr7H2Di9)YsMd2okMI1eiJLH9(q0W)H8YrnmP6Wr90BaIcxp(6WjUK0WopCuqm6gG7w3rXiLaKACuADi9gJpuYjFOHjvhzoy7qVnsMny5q(pu6oKMhA3HW93OW(gchFiACip1rnmP6WrXbBh6TrCIljnKZdhfeJUb4U1DumsjaPghLgCiToKEJXhk5Kp0WKQJmhSDO3gjZgSCi)hkDhsZdLCYhc3FJc7BiC8Hw5qRYrnmP6WrH9n8EJsVnItCj9QCE4OGy0na3TUJQFCuyqCudtQoCujgsn6gWrLymFWrrRxoQedrfJfCulfblkHozcrjLfCItCucPI3GGDE4sIwNhokigDdWDR7OgMuD4OW(gEVb4QMOR6fL0eleIJIrkbi14OyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8drJd5LxoQySGJc7B49gGRAIUQxustSqioXLKxopCuqm6gG7w3rXiLaKACuYyGqYCW2rXuSoWF7JuDKHy0na)q7oeRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGFiACiVsNJAys1HJIngJAys1HYuyXrzkSOIXcokFpkHuXBStCjTkNhokigDdWDR7O4aMrQhP6WrLExwaMGpK4BKdjKjbyoe20Bm0pK0hsgskqoebwD)IahA48sQogJ3hcdpdze4q(MGBQiLJAys1HJIngJAys1HYuyXrzkSOIXcokSP3OesfVbb7exsEQZdhfeJUb4U1DudtQoCuDcGSy6nvKsnrzhfBsbokgPeGuJJ6bKmhSDumLqNmHKhMujahvmwWr1jaYIP3urk1eLDuSjf4exsA05HJcIr3aC36okgPeGuJJsiv8gKSqB23GvFmO0)llhA3HEajZbBhftj0jti5HjvcWrnmP6WrjKkEdcToXL0Q58WrbXOBaUBDhfJucqQXrjKkEdsw8k7BWQpgu6)LLdT7qpGK5GTJIPe6KjK8WKkb4OgMuD4OesfVbXlN4ssd78WrbXOBaUBDhfJucqQXrjLfo0khkXqQr3a5LIGfLqNmHOKYchA3HyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMa8dTYH8kDoQHjvhok2ymQHjvhktHfhLPWIkgl4OE(eqXh7KcucPI3yN4eh1ZNak(yNuGsiv8g78WLeTopCuqm6gG7w3rfJfCuCcm8LIaQeagdgh1WKQdhfNadFPiGkbGXGXjUK8Y5HJcIr3aC36oQySGJc3FJQsfLaeh1WKQdhfU)gvLkkbioXL0QCE4OGy0na3TUJAys1HJkLH(JpvVOgmUSLzKQdhfJucqQXrnmPsakiaBb4d5)q06OIXcoQug6p(u9IAW4YwMrQoCIljp15HJcIr3aC36oQySGJIpK32UdfhyVvpFHaygemWrnmP6WrXhYBB3HIdS3QNVqamdcg4exsA05HJcIr3aC36oQySGJc07a3FJkrHbh1WKQdhfO3bU)gvIcdoXL0Q58WrbXOBaUBDhvmwWr9dMVPcGRszgEnstWkSVH92ayh1WKQdh1py(MkaUkLz41inbRW(g2BdGDItCuDkg35HljADE4OgMuD4O0bcgiVRiLJcIr3aC36oXLKxopCudtQoCu6MU5QLpHUJcIr3aC36oXL0QCE4OgMuD4OwkcOB6M7OGy0na3TUtCj5PopCudtQoCuFmOkbSyhfeJUb4U1DItCIJkbqWvhUK8kDErB60q06LJAZqIksHDuVI03lusP)KE108HoKh(Gdv2NMihAPjhAnSP3OesfVbbV2HiWQ7xeGFiCBHdnFPTJa8dX8nrkaNVTVUc4q0sZhIM6ibqeGFiQYsthctpKz)Hw9hs6d96)CiELOWvhhQFaYin5qALMMhslV2Rz(2(6kGd5fnFiAQJeara(HOklnDim9qM9hA1FiPp0R)ZH4vIcxDCO(biJ0KdPvAAEiT8AVM5B7RRao0QO5drtDKaicWpevzPPdHPhYS)qR(dj9HE9FoeVsu4QJd1pazKMCiTstZdP1Q2Rz(2EBFfPVxOKs)j9QP5dDip8bhQSpnro0sto0ASobetiQrVmLqFTdrGv3Via)q42chA(sBhb4hI5BIuaoFBFDfWHOLMpen1rcGia)qRH7VrVcE(LRDiPp0A4(B0RGNFzgIr3a81oKw0UxZ8T91vahYlA(q0uhjaIa8dTgU)g9k45xU2HK(qRH7VrVcE(LzigDdWx7qAr7EnZ32xxbCOvrZhIM6ibqeGFO1W93Oxbp)Y1oK0hAnC)n6vWZVmdXOBa(AhAKdLEFLV(qAr7EnZ32xxbCipLMpen1rcGia)qRH7VrVcE(LRDiPp0A4(B0RGNFzgIr3a81oKwETxZ8T91vahsJ08HOPosaeb4hAnC)n6vWZVCTdj9Hwd3FJEf88lZqm6gGV2H0Av71mFBFDfWHwnA(q0uhjaIa8dTgU)g9k45xU2HK(qRH7VrVcE(LzigDdWx7qJCO07R81hslA3Rz(2(6kGdPHP5drtDKaicWp0A4(B0RGNF5Ahs6dTgU)g9k45xMHy0naFTdPfT71mFBFDfWH0q08HOPosaeb4hAnC)n6vWZVCTdj9Hwd3FJEf88lZqm6gGV2Hg5qP3x5RpKw0UxZ8T92(ksFVqjL(t6vtZh6qE4douzFAICOLMCO1e6KjefgK)ZAhIaRUFra(HWTfo08L2ocWpeZ3ePaC(2(6kGdTkA(q0uhjaIa8dTg5hWstsb5xU2HK(qRr(bS0Kuq(LzigDdWx7qAr7EnZ32B7Ri99cLu6pPxnnFOd5Hp4qL9PjYHwAYHwJdlZ3iRDicS6(fb4hc3w4qZxA7ia)qmFtKcW5B7RRaoKgMMpen1rcGia)qRH7VrVcE(LRDiPp0A4(B0RGNFzgIr3a81oKwRAVM5B7RRao0RIMpen1rcGia)qRH7VrVcE(LRDiPp0A4(B0RGNFzgIr3a81oKw0UxZ8T91vahIwAP5drtDKaicWp0AKFalnjfKF5Ahs6dTg5hWstsb5xMHy0naFTdPfT71mFBFDfWHO1lA(q0uhjaIa8dTg5hWstsb5xU2HK(qRr(bS0Kuq(LzigDdWx7qJCO07R81hslA3Rz(2(6kGdr7QrZhIM6ibqeGFO1i)awAski)Y1oK0hAnYpGLMKcYVmdXOBa(AhslA3Rz(2(6kGdrRgMMpen1rcGia)qRr(bS0Kuq(LRDiPp0AKFalnjfKFzgIr3a81o0ihk9(kF9H0I29AMVTVUc4q0QHO5drtDKaicWp0AKFalnjfKF5Ahs6dTg5hWstsb5xMHy0naFTdPfT71mFBVTVI03lusP)KE108HoKh(Gdv2NMihAPjhAnHuXBqWRDicS6(fb4hc3w4qZxA7ia)qmFtKcW5B7RRaoKgP5drtDKaicWp0AcPI3GKPn)Y1oK0hAnHuXBqYcT5xU2H0I29AMVTVUc4qRgnFiAQJeara(Hwtiv8gKSx5xU2HK(qRjKkEdsw8k)Y1oKw0UxZ8T92M(Tpnra(HEvhAys1XHmfwW5BRJA(IVM4OCupKEPmGJYt8Kd9IGTJ20POFOxXqmn79T1t8Kd5tKhmnNwAPkX3xpZABA4Y(nJuDWiZIKgUSS0UTEINCO03p1hlhYRv59H8kDEr7T926jEYHOjFtKcW08T1t8Kd55dr9agZHEDZENVTEINCipFOxzyOFicWARfc(HErW2HEBKd9qapZAR(ihQwoujhQWhQcSmHCiTAYH8neoBWYHwAYH0BmgWAMVTEINCipFO0ZEdqoev94RJdngtVb4h6HaEM1w9roK0h6H0SdvbwMqo0lc2o0BJKVTEINCipFO0ZePNhsgdeYHQqac5)i5BRN4jhYZhk9LOl(HOw3ZRKESF1hc)m2dTXhehIE)xJahkA5qJE)Ldj9HWFRTJdnhYd6KjK8T1t8Kd55d9cZayFmYSiP9cVnJug4quTjbeYHytWaJQwoeZ3ePa(HK(qviaH8FevTKVT3wp5qP39a7la)q6WstGdXAR(ihshsvboFO0hJbpc(qrhE23qSlFZHgMuDGpuhg65B7WKQdC(HaS2QpI2(tBbmkEBRyKQdVRf)szHvs3on4bK8yQeWTDys1bo)qawB1hrB)PH)wBhQhqUTdtQoW5hcWAR(iA7pTpguLawVJXc(L2cQErz7alK(JvSoWc5ZKQd8TDys1bo)qawB1hrB)P9XGQeW6DmwWpUnW4dRWaJaIsaMVOwDF42omP6aNFiaRT6JOT)0wma2hJmlYTDys1bo)qawB1hrB)P90BaIcxp(6W7AXV(FzjVPmCvzFWzSmS3Rq7o9)YsMd2okMI1eiJLH9Mg(962omP6aNFiaRT6JOT)04GTd92iExl(1sVX4KtEys1rMd2o0BJKzdw8Non3H7VrH9neoMgE6TDys1bo)qawB1hrB)PH9n8EJsVnI31IFnql9gJto5Hjvhzoy7qVnsMnyXF60m5KX93OW(gchVYQUTdtQoW5hcWAR(iA7pTedPgDd4DmwW)srWIsOtMquszbV7h)yq8oXy(GFA962EB9KdLE3dSVa8dbjac9djLfoK4do0WKMCOcFOjXuMr3a5B7WKQdSF8dymktZEFBhMuDG12FASXyulGX3peGCBhMuDG12FAZEqjngFBhMuDG12FACir)jk7KQy32HjvhyT9NgBmg1WKQdLPWI3Xyb)cPI3GGVTdtQoWA7pnE3wVRf)eyHayFJUbUTdtQoWA7pn2ymQHjvhktHfVJXc(zDciMquJEzkHU3yHumXpTExl(X93OxbpNI0javfjQunzKQJKtg3FJEf88sbgUQxu6MgJBlo5KX93OxbpZAR(iklWlzKQJKtM1jGycjhaJ0MMWVTdtQoWA7pn2ymQHjvhktHfVJXc(f6KjefgK)ZTDys1bwB)P90s1H31IFTyDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMaCAqB62jLfwjXqQr3a5LIGfLqNmHOKYcEM20LCY4(B0RGNjWsfax9mMrGDSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pataonwLgsZB7WKQdS2(tJngJAys1HYuyX7ySG)ofJ7nwift8tR31I)hqYCW2rXucDYesEysLaUTdtQoWA7pnoy7OykSqGiL4Z7AXVwAa5hWstsb5nLzHaCScxPkJQxu4)dqQMOWFRTJksTJ1TH3BIm(BTDO4GTJIPe6KjKmbStfyfS)bycWx5vPzYjR1dizoy7OykHozcjpmPsa7EajZbBhftj0jtizcyNkW0y1EbPy8SD2R5TDys1bwB)PXgJrXjWWXYyEdeS31IFw3gEVjY4V12HId2okMsOtMqYeWovGvW(hGjaFfVsNN14lqdi)awAskiVPmleGJv4kvzu9Ic)Fas1ef(BTDurQB7WKQdS2(t7P3aefUE81H31IF9)YsEtz4QY(GZyzyVxH2D6)LLmhSDumfRjqgld7nnw1TDys1bwB)PPxgaZ6pjfO0BRoqWExl(1)llzHozcjZ7nXow3gEVjY4V12HId2okMsOtMqYeWovGvW(hGjaFfnEBhMuDG12FAL9X04QdVRf)dtQeGccWwaEfA1wlAFbYyGqY4HrQLIbCfU)gCgIr3aCn3P)xwYBkdxv2hCgld79k(xTD6)LLSqNmHK59MyhRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGVIgVTdtQoWA7pTY(yAC1H31I)HjvcqbbylaVIx70)ll5nLHRk7doJLH9Ef)R2o9)YswOtMqY8EtSJ1TH3BIm(BTDO4GTJIPe6KjKmbStfyfS)bycWxrJ70aYpGLMKcYL9X04kbOEAbcPgZoT0azmqi5fsBvIpqH9n8EdodXOBaEYjR)xwYlK2QeFGc7B49gC()O5TDys1bwB)Pv2htJRo8Uw8pmPsakiaBb4v8AN(FzjVPmCvzFWzSmS3R4F12P)xwYL9X04kbOEAbcPgtMa2Pcmn8Ah5hWstsb5Y(yACLaupTaHuJ52omP6aRT)0k7JPXvhExl(1)ll5nLHRk7doJLH9Ef)061ozmqizC)nkwh8FjzigDdW3jJbcjVqARs8bkSVH3BWzigDdW3r(bS0KuqUSpMgxja1tlqi1y2P)xwYcDYesM3BIDSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pata(kA82omP6aRT)0srkBxeqTaMu)HW9Uw8R3y8oPSGsAfVaASQ0DBhMuDG12FA4V12HkrzGLccU31IF9gJ3jLfusR4fqdV0q32HjvhyT9Ng(BTDO4GTJIPe6KjeVRf)6ngVtklOKwXlGg0QXB7WKQdS2(tZ3eCvVOs9n8j8Uw8J7VrH9neUFnEB9KdL(xo0lsGHJLX8gi4dne4qJHadN(HgMujaVpu0hkaGFiPpeEsahc7BiC8TDys1bwB)P5BcUQxuP(g(eExl(X93OW(gcFf)RANwpGK5ey4yzmVbsEysLaso5hqYCW2rXucDYesEysLa082omP6aRT)08nbx1lQuFdFcVRf)4(BuyFdHVIFA3P)xwYbi(aI6PjYyY)NDSUn8EtKzJXO4ey4yzmVbcota7ubEfVEbPy8SD2FBhMuDG12FA(MGR6fvQVHpH31IFC)nkSVHWxXpT7yDB49MiJ)wBhkoy7OykHozcjta7ubwb7FaMaCAKIXZ2z)oPSWk06LNtX4z7SFNw6)LLmNadhlJ5nqW5)Zo9)YsMtGHJLX8gi4mbStf4vgMuDK9nbx1lQuFdFImShyFbuszbThMuDKXFRTdfhSDumLqNmHKH9a7lGsklO5oT0azmqiz83A7qLOmWsbbpdXOBaEYjR)xwYjkdSuqWZ)hnVTdtQoWA7pn2ymQHjvhktHfVJXc(zDciMquJEzkHU3yHumXpTExl(1awNaIjKCcieF0j3wp5qVIs81F5qudJulfd4hIQ)gS3hIQ)MdrjK6nCOcFiSq6ifqoK4BId9IGTd92iEFiCFOsoKVbFO5q(Qu(aYHEivtkH(TDys1bwB)PH7VrHfs9g8Uw8RbYyGqY4HrQLIbCfU)gCgIr3a8BRNCiQhi4h6fbBhf7q0uta8HwAYHO6V5qu(gchFOFiL5qEqNmHCiw3gEVjouHpeZ0y4qsFicmC632HjvhyT9NghSDO3gX7AXV(FzjZbBhftXAcKjWWKD4(BuyFdHtdpDhRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGVIxP726jh6f)KksDipOtMqoegK)J3hc)ab)qViy7OyhIMAcGp0stoev)nhIY3q44B7WKQdS2(tJd2o0BJ4DT4x)VSK5GTJIPynbYeyyYoC)nkSVHWPHNUJ1TH3BIm(BTDO4GTJIPe6KjKmbStfyfS)bycWPbTEDBhMuDG12FACW2HEBeVRf)6)LLmhSDumfRjqMadt2H7VrH9neon80DAP)xwYCW2rXuSMazSmS3R4vYjlJbcjJhgPwkgWv4(BWzigDdW182omP6aRT)04GTd92iExl(1)llzoy7OykwtGmbgMSd3FJc7BiCA4P7gMujafeGTa8k0EBhMuDG12FA4(BuyHuVHB7WKQdS2(tJngJAys1HYuyX7ySGFwNaIje1OxMsOFB9KdL(xoe9(Fi2ehkfihsFyVpK0hsJhIQ)Mdr5BiC8H0HLMah6fjWWXYyEde8HyDB49M4qf(qey409(qLSg(q97H(HK(q4hi4hs8b2df9MB7WKQdS2(tZ3eCvVOs9n8j8Uw8J7VrH9ne(k(x1ow3gEVjY4V12HId2okMsOtMqYeWovGvW(hGjaFfV04oTKXaHK5GTJIPyJXurQmeJUb4jNmRBdV3ez2ymkobgowgZBGGZeWovGxrlT0ONX93OW(gcxZxWWKQJm23W7nk92izypW(cOKYcAQ9WKQJSVj4QErL6B4tKH9a7lGsklO5TDys1bwB)PX726nJoZakziPab7NwVRf)eyHayFJUb2jLfwjXqQr3a5LIGfLqNmHOKYc32HjvhyT9Ng23W7nk92i32B7WKQdCgB6nkHuXBqW()yqvcy9ogl4h3FJbePIukYxNU31IFw3gEVjY4V12HId2okMsOtMqYeWovGvW(hGjaNgYqsbsMxyzcgS614oPSWkjgsn6giVueSOe6KjeLuwWZAjdjfizEHLjyWQxJAEBhMuDGZytVrjKkEdcwB)P9XGQeW6DmwWp(h6MU5QXcIp6yX7AXpRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGtdziPajZlSmbdw9ACNuwyLedPgDdKxkcwucDYeIskl4zTKHKcKmVWYemy1RrnVTEYHELeEmbdoKVbFO5q061HWaRd(H4GzOFOj4hQWhs8beyPjWHWVRNhGFOLMCOLIGLd5bDYeYHK(qMkGd9Fo0Ms8DiXhCicGLB7WKQdCgB6nkHuXBqWA7pTpguLawVJXc(b7dDcmgvt4XemW7AXpRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGtdTKHKcKmVWYemy1Rrn1MwV2X62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4ROLwAjdjfizEHLjyWQxJAQnTEPPNPvJAUtklSsIHuJUbYlfblkHozcrjLf8SwAjdjfizEHLjyWQxJAQnTEP5T92omP6aNzDciMquJEzkHUFC)nkslExl(X93OxbpNI0javfjQunzKQJDAX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb40WR0LCYSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pata(kRkDAEBhMuDGZSobetiQrVmLqxB)PH7VrrAX7AXpU)g9k45LcmCvVO0nng3w8UhqYCW2rXucDYesEysLaUTdtQoWzwNaIje1OxMsORT)0W93OiT4DT4h3FJEf88MYWv((HOKHjfdFBhMuDGZSobetiQrVmLqxB)PXbwzhPIuk92iExl(1c3FJEf8SbgUsNUc2p2hdKCY4(B0RGNFdjQaR6o9ayQiLM706bKmhSDumLqNmHKhMujGD4(BuyFdHtdVsozn4bKmhSDumLqNmHKhMujGDSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pata(kEA6082omP6aNzDciMquJEzkHU2(tJdSYosfPu6Tr8Uw8RfU)g9k45LMKc0BsakcKaifGtozTW93OxbpNOnJugqHBtciKDAaU)g9k453qIkWQUtpaMksPPM70GhqYCW2rXucDYesEysLaUTdtQoWzwNaIje1OxMsORT)0wma2hJmlI31IFC)n6vWZjAZiLbu42KacX7keGq(pIQw8R)xwYjAZiLbu42Kacj)FUTdtQoWzwNaIje1OxMsORT)0WS(tQiLskXh4DT4h3FJEf8mRT6JOSaVKrQo29asMd2okMsOtMqYdtQeWTDys1boZ6eqmHOg9YucDT9NgM1FsfPusj(aVRf)AaU)g9k4zwB1hrzbEjJuDCBhMuDGZSobetiQrVmLqxB)Pv2hi4vKsXgzWcPF8bExl(FajZbBhftj0jti5HjvcyhU)gf23q4(t3T92omP6aN99OesfVX()yqvcy9ogl4hxXY3OszgEnstWkWQBa7TDys1bo77rjKkEJ12FAFmOkbSEhJf8JRy5Bud(Pitiyfy1nG92EBhMuDGZDkg3VoqWa5DfPUTdtQoW5ofJRT)00nDZvlFc9B7WKQdCUtX4A7pTLIa6MU532Hjvh4CNIX12FAFmOkbS4B7TDys1bo)8jGIp2jfOesfVX()yqvcy9ogl4NtGHVueqLaWyWCBhMuDGZpFcO4JDsbkHuXBS2(t7JbvjG17ySGFC)nQkvucqUTdtQoW5Npbu8XoPaLqQ4nwB)P9XGQeW6DmwWFkd9hFQErnyCzlZivhExl(hMujafeGTaSFAVTdtQoW5Npbu8XoPaLqQ4nwB)P9XGQeW6DmwWpFiVTDhkoWERE(cbWmiyWTDys1bo)8jGIp2jfOesfVXA7pTpguLawVJXc(b9oW93Osuy42omP6aNF(eqXh7KcucPI3yT9N2hdQsaR3Xyb))G5BQa4QuMHxJ0eSc7ByVna(2EBhMuDGZcPI3GG9)XGQeW6DmwWp23W7nax1eDvVOKMyHq8Uw8Z62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb40WlVUTdtQoWzHuXBqWA7pn2ymQHjvhktHfVJXc(99OesfVXExl(LXaHK5GTJIPyDG)2hP6idXOBa(ow3gEVjY4V12HId2okMsOtMqYeWovGvW(hGjaNgELUBRNCO07YcWe8HeFJCiHmjaZHWMEJH(HK(qYqsbYHiWQ7xe4qdNxs1Xy8(qy4ziJahY3eCtfPUTdtQoWzHuXBqWA7pn2ymQHjvhktHfVJXc(XMEJsiv8ge8TDys1bolKkEdcwB)P9XGQeW6DmwWFNailMEtfPutu2rXMuG31I)hqYCW2rXucDYesEysLaUTdtQoWzHuXBqWA7pnHuXBqO17AXVqQ4nizAZ(gS6JbL(Fzz3dizoy7OykHozcjpmPsa32Hjvh4SqQ4niyT9NMqQ4niE5DT4xiv8gKSxzFdw9XGs)VSS7bKmhSDumLqNmHKhMujGB7WKQdCwiv8geS2(tJngJAys1HYuyX7ySG)Npbu8XoPaLqQ4n27AXVuwyLedPgDdKxkcwucDYeIsklSJ1TH3BIm(BTDO4GTJIPe6KjKmbStfyfS)bycWxXR0DBVTdtQoWzHozcrHb5)4paXhqupnrgJ31IFw3gEVjY4V12HId2okMsOtMqYeWovGvW(hGjaNg0QXB7WKQdCwOtMquyq(pA7pTuKY2fbulGj1FiCVRf)SUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pataonOvd7zTgMuDKXFRTdfhSDumLqNmHKH9a7lGsklO9WKQJm23W7nk92izypW(cOKYcAUtlw3gEVjYSXyuCcmCSmM3abNjGDQatdA1WEwRHjvhz83A7qXbBhftj0jtizypW(cOKYcApmP6iJ)wBhQeLbwki4zypW(cOKYcApmP6iJ9n8EJsVnsg2dSVakPSGMjN8dizobgowgZBGKjGDQaVcRBdV3ez83A7qXbBhftj0jtizcyNkWky)dWeGR9WKQJm(BTDO4GTJIPe6KjKmShyFbuszbnVTdtQoWzHozcrHb5)OT)0WFRTdvIYalfeCVRf)AX62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb40Gwn6zTgMuDKXFRTdfhSDumLqNmHKH9a7lGsklO5oTyDB49MiZgJrXjWWXYyEdeCMa2PcmnOvJEwRHjvhz83A7qXbBhftj0jtizypW(cOKYcApmP6iJ)wBhQeLbwki4zypW(cOKYcAMCYpGK5ey4yzmVbsMa2Pc8kSUn8EtKXFRTdfhSDumLqNmHKjGDQaRG9pataU2dtQoY4V12HId2okMsOtMqYWEG9fqjLf0uZKtwlnG8dyPjPG8MYSqaowHRuLr1lk8)bivtu4V12rfP2X62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb4R4PPtZB7WKQdCwOtMquyq(pA7pn2ymkobgowgZBGG9Uw8Z62W7nrg)T2ouCW2rXucDYesMa2PcSc2)amb40GwV8SwdtQoY4V12HId2okMsOtMqYWEG9fqjLf0Eys1rg7B49gLEBKmShyFbuszbn3jLfwjXqQr3a5LIGfLqNmHOKYcEMwV88WKQJmBmgfNadhlJ5nqWzypW(cOKYcApmP6iJ)wBhkoy7OykHozcjd7b2xaLuwq7HjvhzSVH3Bu6TrYWEG9fqjLfUTdtQoWzHozcrHb5)OT)0WFRTdfhSDumLqNmH4DT4xklSsIHuJUbYlfblkHozcrjLf2P1dizobgowgZBGKhMujGDpGK5ey4yzmVbsMa2Pc8kdtQoY4V12HId2okMsOtMqYWEG9fqjLf0CNwAGmgiKm(BTDOsugyPGGNHy0nap5KFajNOmWsbbppmPsaAUtlC)nkSVHW9NUKtwRhqYCcmCSmM3ajpmPsa7EajZjWWXYyEdKmbStfyAmmP6iJ)wBhkoy7OykHozcjd7b2xaLuwq7HjvhzSVH3Bu6TrYWEG9fqjLf0m5K16bKCIYalfe88WKkbS7bKCIYalfe8mbStfyAmmP6iJ)wBhkoy7OykHozcjd7b2xaLuwq7HjvhzSVH3Bu6TrYWEG9fqjLf0m5K1s)VSKtrkBxeqTaMu)HWZ)ND6)LLCksz7IaQfWK6peEMa2PcmngMuDKXFRTdfhSDumLqNmHKH9a7lGsklO9WKQJm23W7nk92izypW(cOKYcAQPtCIZba]] )


end
